#!/usr/bin/env python

from __future__ import print_function
import parser as symbolizer

__doc__ = """
Usage:
  symbolizer (-c CONFIG.yml | DIR [DIR ...]
                 [-r ROOTDIR] [-o OUTPUT]
    [--numeric] [--mixedcase] [--uppercase]
                    [-S suf,fix,es,etc ...]
                    [-R r,r0,r1,r2,etc ...])
             [-V | --verbose]
  symbolizer  -h | --help
  symbolizer  -v | --version
    
Options:
  -c CONFIG.yml --config=CONFIG.yml    read options from CONFIG.yml (overrides CLI)
  -r ROOTDIR --root=ROOTDIR            root dir for resolving paths [default: CWD]
  -o OUTPUT --output=OUTPUT            output file [default: stdout]
  --numeric                            symbolize 8754 (numeric) tokens [default: false]
  --mixedcase                          symbolize Mixed_Case_Tokens [default: false]
  --uppercase                          symbolize UPPER_CASE_TOKENS [default: false]
  -S suf,fix,es --suffixes=suf,fix,es  file suffixes to parse [default: %s]
  -R r,r0,r1,r2 --reserved=r,r0,r1,r2  extra reserved words (DON'T symbolize these)
  -V --verbose                         print verbose output
  -h --help                            show this text
  -v --version                         print version

""" % ", ".join(tuple(symbolizer.SUFFIXES))

from os.path import join, isdir, dirname
from docopt import docopt
import templates
import sys

__version__ = "0.0.1"
exec(compile(open(join(dirname(__file__), 'version.py')).read(),
             'version.py', 'exec'))

def cli(argv=None):
    if not argv:
        argv = sys.argv
    
    arguments = docopt(__doc__, argv=argv[1:],
                                help=True,
                                version=__version__)
    
    # print(argv)
    # print(arguments)
    # sys.exit()
    
    verbose = arguments.pop('--verbose', False)
    output = arguments.pop('--output')
    
    directories = arguments.pop('DIR', [])
    if not directories:
        print("ERROR: Need to specify a directory to scan",
            file=sys.stderr)
        print(__doc__)
    
    # check directories
    for directory in directories:
        if not isdir(directory):
            print("ERROR: not a directory:",
                file=sys.stderr)
            print("\t%s" % directory,
                file=sys.stderr)
    
    # scan directories
    symbols = set()
    collected = set()
    for directory in directories:
        if verbose:
            print("> Scanning directory: %s" % directory,
                file=sys.stderr)
        collected |= symbolizer.collect(directory)
    
    # parse collected source files
    for source_file in sorted(collected):
        if verbose:
            print("> Parsing source: %s" % source_file,
                file=sys.stderr)
        symbols |= symbolizer.parse_source_file(source_file)
    
    # generate header
    if verbose:
        print("> Generating header:",
                file=sys.stderr)
    if output == "stdout":
        print()
        print(templates.generate_header(symbols))
        print()
    else:
        if verbose:
            print("> Writing output: %s" % output,
                file=sys.stderr)
        with open(output, "wb") as fh:
            fh.write(templates.generate_header(symbols))
    
    # donezo.
    if verbose:
        print("> Done.")


if __name__ == '__main__':
    cli(sys.argv)


from pkgutil import extend_path
__path__ = extend_path(__path__, __name__)

from parser import (
    SUFFIXES, KEYWORDS, SYMBOL_PREFIX,
    sanitizer, symbol_re, suffix_re,
    quotes, dbl_quotes, cpp_comments, c_comments, blockout,
    collect_files, sanitize_suffixes, collect,
    parse, parse_source_file)

from templates import generate_header


#!/usr/bin/env python

from __future__ import print_function

from collections import defaultdict, namedtuple
# from operator import itemgetter

def autoincrementer():
    """ Get an independantly auto-incrementing unary function """
    def inc():
        """ Get a uniquely auto-incremented integer value """
        if not hasattr(inc, 'v'):
            setattr(inc,    'v', 0)
        out = getattr(inc,  'v')
        setattr(inc,        'v', out + 1)
        return out
    return inc

class AutoInc(object):
    # Get a uniquely auto-incremented integer value
    __slots__ = ('v',)
    
    def __init__(self, value=0):
        try:
            # assume we got another AutoInc ...
            # ... I could sure use SFINAE right about now
            self.v = value.v
        except AttributeError:
            self.v = int(value)
    
    def __call__(self):
        self.v += 1
        return self.v
    
    def add(self, value):
        self.v += int(value)
    
    def __int__(self):
        return self.v
    def __float__(self):
        return float(self.v)
    def __long__(self):
        return long(self.v)
    
    def __repr__(self):
        return str(self.v)

class Histogram(defaultdict):
    
    @staticmethod
    def lexcmp(k0, k1):
        """ Hacky comparator to privilege string length """
        return cmp(len(k1), len(k0)) or cmp(k0, k1)
    
    @staticmethod
    def autoinc():
        return AutoInc()
    
    def __init__(self, *args, **kwargs):
        if len(args) and not len(kwargs):
            for arg in args:
                argcls_name = arg.__class__.__name__
                if argcls_name.startswith('FrozenHistogram'):
                    # reconstitute FrozenHistogram
                    kwargs.update(arg._asdict())
                else:
                    # normal arg: column name
                    kwargs.update({ arg: AutoInc() })
        super(Histogram, self).__init__(
            Histogram.autoinc, *tuple(), **kwargs)
    
    def prettyprint(self, **kwargs):
        from pprint import pformat
        return "%s%s" % (
            self.__class__.__name__,
            pformat((Histogram.autoinc, dict(self)), **kwargs))
    
    def __repr__(self):
        return self.prettyprint()
    
    def __str__(self):
        longest_key = sorted(self.iterkeys(), cmp=Histogram.lexcmp)
        return self.prettyprint(
            depth=10,
            width=len(longest_key)/2)
    
    ### SET LOGIC ###
    def setresult(self, keyset):
        pass
    
    def __and__(self, rhs):
        """ & (intersection) """
        pass
    
    def __or__(self, rhs):
        """ | (union) """
        pass
    
    def __xor__(self, rhs):
        """ ^ (symmetric difference) """
        pass
    
    def __sub__(self, rhs):
        """ - (difference [NON-COMMUTATIVE!]) """
        pass
    
    def inc(self, idx):
        return self[str(idx)]()
    
    def add(self, idx, value):
        return self[str(idx)].add(value)
    
    def keyset_hash(self):
        return hex(abs(hash(self.iterkeys())))
    
    def frozentype(self, rename=True):
        return namedtuple('FrozenHistogram_%s' % self.keyset_hash(),
            tuple(self.iterkeys()), rename=rename)
    
    def freeze(self):
        return self.frozentype()(**dict(
            (ituple[0], int(ituple[1])) for ituple in self.iteritems()
        ))
    
    def normalize(self):
        ceil = float(sum(int(v) for v in self.itervalues()))
        return self.frozentype()(**dict(
            (ituple[0], float(ituple[1]) / ceil) for ituple in self.iteritems()
        ))
    
    def max(self):
        return max(*(int(v) for v in self.values()))
    def min(self):
        return min(*(int(v) for v in self.values()))





from datetime import datetime

header_start_tpl = """
///
/// Generated by iod-symbolizer at %s
/// iod-symbolizer by Alexander Bohn -- http://github.com/fish2000/iod-symbolizer
/// IOD library by Matthieu Garrigues -- https://github.com/matt-42/iod
///

#include <iod/symbol.hh>

"""

header_num_symbol_tpl = """
#ifndef IOD_SYMBOL_%(symbol)s
    #define IOD_SYMBOL_%(symbol)s
    iod_define_number_symbol(%(symbol)s)
#endif

"""

header_std_symbol_tpl = """
#ifndef IOD_SYMBOL_%(symbol)s
    #define IOD_SYMBOL_%(symbol)s
    iod_define_symbol(%(symbol)s)
#endif

"""

def generate_header(symbol_set, when=None):
    if when is None:
        when is datetime.now().isoformat()
    
    header_out = header_start_tpl % when
    
    for symbol in symbol_set:
        try:
            int(symbol)
        except ValueError:
            # it's not a number
            header_out += header_std_symbol_tpl % dict(symbol=symbol)
        else:
            # ok, it's a number
            header_out += header_num_symbol_tpl % dict(symbol=symbol)
    
    return header_out


#!/usr/bin/env python
# encoding: UTF-8

from __future__ import print_function

import unittest
import symbolizer

sample_input = unicode("""
// Copyright 2012-2014 Alexander Böhn <fish2000@gmail.com>
// License: MIT (see COPYING.MIT file)

#ifndef LIBIMREAD_IO_PNG_HH_
#define LIBIMREAD_IO_PNG_HH_

#include <cstring>
#include <csetjmp>
#include <vector>
#include <sstream>

#include <libimread/libimread.hpp>
#include <libimread/base.hh>
#include <libimread/options.hh>

namespace im {
    
    /*
    using namespace symbols::s;
    
    auto options =
    D(
        _compression_level     = -1,
        _backend               = "io_png"
    );
    */
    
    namespace PNG {
        
        struct format {
            
            static auto opts_init() {
                return D(
                    _signature = "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"
                );
            }
            
            using options_type = decltype(format::opts_init());
            static const options_type options;
        };
        
    }
    
    #define PNG_SIGNATURE "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"
    
    // namespace signature {
    //     struct PNGSignature {
    //         PNGSignature() {}
    //         operator char const*() { return "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"; }
    //     };
    //     static const PNGSignature PNG;
    // }
    
    class PNGFormat : public ImageFormat {
        public:
            typedef std::true_type can_read;
            typedef std::true_type can_write;
            
            /// NOT AN OVERRIDE:
            static bool match_format(byte_source *src) {
                return match_magic(src, PNG_SIGNATURE, 8);
            }
            
            virtual std::unique_ptr<Image> read(byte_source *src,
                                                ImageFactory *factory,
                                                const options_map &opts) override;
            virtual void write(Image &input,
                               byte_sink *output,
                               const options_map &opts) override;
    };
    
    namespace format {
        using PNG = PNGFormat;
    }
    
}


#endif /// LIBIMREAD_IO_PNG_HH_
""", encoding='UTF-8', errors="replace").encode('UTF-8', errors="replace").strip()

sample_output = set(['compression_level', 'backend', 'signature'])

sample_header = """
///
/// Generated by iod-symbolizer at RIGHT NOW
/// iod-symbolizer by Alexander Bohn -- http://github.com/fish2000/iod-symbolizer
/// IOD library by Matthieu Garrigues -- https://github.com/matt-42/iod
///

#include <iod/symbol.hh>


#ifndef IOD_SYMBOL_signature
    #define IOD_SYMBOL_signature
    iod_define_symbol("signature")
#endif


#ifndef IOD_SYMBOL_compression_level
    #define IOD_SYMBOL_compression_level
    iod_define_symbol("compression_level")
#endif


#ifndef IOD_SYMBOL_backend
    #define IOD_SYMBOL_backend
    iod_define_symbol("backend")
#endif

"""

class SymbolizerTests(unittest.TestCase):
    
    def setUp(self):
        self.maxDiff = 1000
    
    def test_symbolize(self):
        source_edit = (sample_input,)
        for blockout_re in (symbolizer.quotes, symbolizer.dbl_quotes,
                            symbolizer.cpp_comments, symbolizer.c_comments):
            source_edit = blockout_re.subn(symbolizer.blockout, source_edit[0])
        try:
            sym = set(symbolizer.symbol_re.findall(source_edit[0])) - symbolizer.KEYWORDS
        except:
            sym = set()
        
        self.assertEqual(sym, sample_output)
    
    def test_symbolize_parse(self):
        self.assertEqual(symbolizer.parse(sample_input), sample_output)
    
    def test_generate_header(self):
        header = symbolizer.generate_header(sample_output, when="RIGHT NOW")
        self.assertEqual(header.strip(), sample_header.strip())
    
    def test_file_read(self):
        pass

sample_config = """

root: ~/Dropbox/libimread

source:
- src
- include/libimread

suffixes:       [c, cc, cpp, cxx, m, mm, h, hh, hpp, hxx, inl]
reserved:       [assert, raise, error, bridge, color_map]

target:
  file:       include/libimread/symbols.hpp
  overwrite:  true

"""

from symbolizer import config
from os.path import join, isdir, expanduser

class ConfigTests(unittest.TestCase):
    
    def test_parse_config(self):
        config_dict = config.parse(sample_config)
        
        root = expanduser(config_dict['root'])
        
        self.assertTrue(isdir(root))
        for sourcedir in config_dict['source']:
            self.assertTrue(isdir(join(root, sourcedir)))
        
        if 'suffixes' in config_dict:
            self.assertNotEqual(len(config_dict['suffixes']), 0)
        if 'reserved' in config_dict:
            self.assertNotEqual(len(config_dict['reserved']), 0)
        
        


if __name__ == '__main__':
    unittest.main()




from __future__ import print_function

import yaml
from os.path import join, isdir, isfile, expanduser

def parse(configuration, check=True):
    config_dict = yaml.safe_load(configuration)
    if not check:
        return config_dict
    
    root = expanduser(config_dict['root'])
    assert isdir(root)
    for sourcedir in config_dict['source']:
        assert isdir(join(root, sourcedir))
    
    if 'suffixes' in config_dict:
        assert len(config_dict['suffixes']) > 0
    if 'reserved' in config_dict:
        assert len(config_dict['reserved']) > 0
    
    target = config_dict.get('target')
    if 'file' in target:
        targfile = join(root, target.get('file'))
        if not target.get('overwrite'):
            if isfile(targfile):
                raise IOError("File output target exists,\n"
                              "but configuration forbids overwriting\n")
    return config_dict




#!/usr/bin/env python

__version__ = "0.0.1"
exec(compile(open('symbolizer/version.py').read(),
             'symbolizer/version.py', 'exec'))

name = 'symbolizer'
long_name = 'iod-symbolizer'
version = __version__
packages = [name]
description = "Parse C/C++/Objective-C/Objective-C++ source for IOD symbols and output header files"

keywords = [
    'IOD','C','C++','Objective-C','Objective-C++','C++11','C++14',
    'Preprocessor','Parser','Symbols','Header'
]

long_description = """
    
    iod-symbolizer - Parse C/C++/Objective-C/Objective-C++ source for IOD symbols and output header files.
    
"""


classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Other Environment',
    'Environment :: Plugins',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Topic :: Database',
    'Topic :: Utilities',
]

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import sys

if 'sdist' in sys.argv and 'upload' in sys.argv:
    """ CLEAN THIS MESS UP RIGHT NOW YOUNG MAN """
    import commands
    import os
    finder = "/usr/bin/find %s \( -name \*.pyc -or -name .DS_Store \) -delete"
    theplace = os.getcwd()
    if theplace not in (".", "/"):
        print("+ Deleting crapola from %s..." % theplace)
        print("$ %s" % finder % theplace)
        commands.getstatusoutput(finder % theplace)
        print("")

setup(

    name=long_name, version=version, description=description,
    long_description=long_description,
    download_url=('http://github.com/fish2000/%s/zipball/master' % long_name),

    author=u"Alexander Bohn",
    author_email='fish2000@gmail.com',
    url='http://bitbucket.org/fish2000/%s' % long_name,
    license='GPLv2',
    keywords=', '.join(keywords),
    platforms=['any'],
    
    include_package_data=True,
    
    packages=[]+packages,
    
    package_dir={
        'symbolizer': 'symbolizer',
    },
    
    entry_points={
        'console_scripts': [
            'symbolizer = symbolizer.cli:cli'
        ],
    },
    
    package_data={},
    install_requires=['docopt', 'PyYAML'],

    classifiers=classifiers+[
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: OS Independent',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: PyPy'],
)




#!/usr/bin/env python

from __future__ import print_function

import re
import functools
from itertools import ifilter, imap
from os.path import join, expanduser, abspath, isdir, isabs
from os import walk, getcwd

SUFFIXES = set((
    r"c",
    r"cc", r"cpp", r"cxx",
    r"m", r"mm",
    r"h",
    r"hh", r"hpp", r"hxx",
    r"inl"
))

KEYWORDS = set((
    "alignas", "alignof", "and", "and_eq", "assert", "asm", "auto",
    "bitand", "bitor", "bool", "break",
    "case", "catch", "char", "char16_t", "char32_t", "class",
    "compl", "const", "constexpr", "const_cast", "continue",
    "decltype", "default", "delete", "do", "double", "dynamic_cast",
    "else", "enum", "explicit", "export", "extern",
    "false", "float", "for", "friend",
    "goto",
    "if", "inline", "int",
    "long",
    "mutable",
    "namespace", "new", "noexcept", "not", "not_eq", "nullptr",
    "operator", "or", "or_eq",
    "private", "protected", "public",
    "register", "reinterpret_cast", "return",
    "short", "signed", "sizeof", "static", "static_assert",
    "static_cast", "struct", "switch",
    "template", "this", "thread_local", "throw", "true",
    "try", "typedef", "typeid", "typename",
    "union", "unsigned", "using",
    "virtual", "void", "volatile",
    "wchar_t", "while",
    "xor", "xor_eq"))

SYMBOL_PREFIX = r'_'

sanitizer = re.compile(r'\W+')
# symbol_re = re.compile(r"(?:\w+)%s([a-zA-Z][0-9a-zA-Z_]+)(?!_t)" % SYMBOL_PREFIX)
symbol_re = re.compile(r"(?:\s+)(?:%s)([0-9a-z][0-9a-z_]+)(?!_t)" % SYMBOL_PREFIX)
suffix_re = re.compile(r"(%s)$" % ("|".join(map(
    lambda suffix: r"\.%s" % suffix, SUFFIXES))))

quotes          = re.compile(r"'(.*)'",     re.MULTILINE)
dbl_quotes      = re.compile(r'"(.*)"',     re.MULTILINE)
cpp_comments    = re.compile(r'//(.*)\n',   re.MULTILINE)
c_comments      = re.compile(r'/\*(.*)\*/', re.MULTILINE)
blockout        = lambda match: "#" * len(match.group(0))

def collect_files(root_dir, suffixes):
    """ Collect files recursively that match one or more file suffixes.
    """
    collected = set()
    suffix_re = re.compile(r"(%s)$" % ("|".join(map(
        lambda suffix: r"\.%s" % suffix, suffixes))))
    
    for pth, dirs, files in walk(expanduser(root_dir)):
        found = set(imap(lambda file: join(pth, file),
            ifilter(lambda file: suffix_re.search(file, re.IGNORECASE), files)))
        collected |= found
    return collected

def sanitize_suffixes(*suffixes):
    """ Ensure suffixes (which may be passed in from the user) are clean
    """
    return set((sanitizer.sub('', suffix.lower()) for suffix in suffixes))

collect = functools.partial(collect_files, suffixes=list(sanitize_suffixes(*SUFFIXES)))

def parse(source):
    source_edit = (source,)
    for blockout_re in (quotes, dbl_quotes, cpp_comments, c_comments):
        source_edit = blockout_re.subn(blockout, source_edit[0])
    try:
        sym = set(symbol_re.findall(source_edit[0]))
    except:
        return set()
    return sym - KEYWORDS

def parse_source_file(file_name):
    with open(file_name, "rb") as fh:
        source_file = fh.read()
    return parse(source_file)

class Parser(object):
    
    quotes          = re.compile(r"'(.*)'",     re.MULTILINE)
    dbl_quotes      = re.compile(r'"(.*)"',     re.MULTILINE)
    cpp_comments    = re.compile(r'//(.*)\n',   re.MULTILINE)
    c_comments      = re.compile(r'/\*(.*)\*/', re.MULTILINE)
    blockout        = lambda match: "#" * len(match.group(0))
    
    def __init__(self, *args, **arguments):
        super(Parser, self).__init__(args, arguments)
        
        self.suffixes = SUFFIXES
        self.keywords = KEYWORDS
        self.dirs = set()
        self.collected = set()
        self.symbols = set()
        self.symbol_re = re.compile(
            r"(?:\s+)(?:%s)([0-9a-z][0-9a-z_]+)(?!_t)" % SYMBOL_PREFIX)
        
        self.verbose = arguments.pop('--verbose', False)
        
        symbolize_opts = dict(
            numeric=arguments.pop('--numeric', False),
            mixedcase=arguments.pop('--mixedcase', False),
            uppercase=arguments.pop('--uppercase', False))
        
        symbol_init_range = r'a-z'
        symbol_main_range = r'0-9a-z_'
        
        # THIS IS SOME HACK SHIT
        if symbolize_opts['numeric']:
            symbol_init_range = r'0-9' + symbol_init_range
        if symbolize_opts['mixedcase'] or symbolize_opts['uppercase']:
            symbol_main_range = r'0-9a-zA-Z_'
            if symbolize_opts['uppercase']:
                # for now 'mixed-case' is implied
                symbol_init_range = symbol_init_range + r'A-Z'
        
        self.symbol_re = re.compile(
            r"(?:\s+)(?:%s)([%s][%s]+)(?!_t)" % (
                SYMBOL_PREFIX,
                symbol_init_range,
                symbol_main_range))
        
        suffixes = arguments.pop('--suffixes', None)
        if suffixes:
            self.suffixes = sanitize_suffixes(suffixes)
        
        self.suffix_re = re.compile(r"(%s)$" % ("|".join(map(
            lambda suffix: r"\.%s" % suffix, self.suffixes))))
        
        root = arguments.pop('--root')
        if root == "CWD":
            root = abspath(getcwd())
        else:
            root = abspath(root)
        if not isdir(root):
            raise IOError("Bad root dir: %s" % root)
        
        for directory in arguments['DIR']:
            if isabs(directory):
                self.dirs.add(directory)
            else:
                dd = join(root, directory)
                if isdir(dd):
                    self.dirs.add(dd)
                else:
                    raise IOError("DIR WTF: %s" % dd)
    
    def collect(self, root_dir):
        for pth, dirs, files in walk(expanduser(root_dir)):
            found = set(imap(lambda file: join(pth, file),
                ifilter(lambda file: self.suffix_re.search(file, re.IGNORECASE), files)))
            self.collected |= found
    
    def collect_all(self):
        for directory in self.dirs:
            self.collect(directory)
    
    def parse(self, source):
        source_edit = (source,)
        for blockout_re in (self.quotes, self.dbl_quotes,
                            self.cpp_comments, self.c_comments):
            source_edit = blockout_re.subn(self.blockout, source_edit[0])
        try:
            sym = set(self.symbol_re.findall(source_edit[0]))
        except:
            sym = set()
        self.symbols |= sym - self.keywords
    
    def parse_source_file(self, file_name):
        with open(file_name, "rb") as fh:
            source_file = fh.read()
        self.parse(source_file)
    
    def parse_collected(self):
        for source_file in self.collected:
            self.parse_source_file(source_file)
    
    def __len__(self):
        return len(self.symbols)
    
    @property
    def directories(self):
        return tuple(self.symbols)
    
    @property
    def files(self):
        return tuple(self.collected)
    
    @property
    def symbols(self):
        return tuple(self.symbols)

